' functions used in all language related modules

#include once "include/TExtendKey.bi"
#include once "include/TExercise.bi"
#include once "include/TInput.bi"
#include once "include/levenshtein.bi"

#define wordtestInfo standardInfo

mdListDeclare(TExercise)
mdMapDeclare(long, long)
mdMapDeclare(string, long)

namespace SETTING_OPTIONS
  const NONE = 0
  const SKIPDIALOG  = 2^0   ' only create exercise.currentList; do not show the setting dialog
  const HASPARENT   = 2^1   ' instruction to show a button for the parent element
  const HASCHILD    = 2^2   ' instruction to show a button for the child element
  const HASPREV     = 2^3   ' instruction to show a button for the previous element
  const HASNEXT     = 2^4   ' instruction to show a button for the following element
  const SHOWTAGS    = 2^5   ' in 'chooseWords': show tag list
  const SHOWVERBREG = 2^6   ' in 'chooseWords': show buttons for regular/irregular verbs
end namespace

namespace SEARCHDIALOG
  type TData
    as integer posOk = 1, posExit = 2, posKeys = 3, w, h, inpX, inpY, inpW
    as string  keyString(any)
  end type
  declare sub incGui(gui as TGui, wdata as TData)
end namespace

enum MENUTYPE
  both = 1        ' left side: exercises with info line; right side: options 
  exercLeft       ' only exercises (left side, with info line)
  exercRight      ' only exercises (right side, no info line)
  exercRightExt   ' only exercises (right side, with info line) 
  exercTwoRows    ' only exercises (left and right side, no info line)
  exercTwoRowsExt ' only exercises (left and right side, with info line)
  exercThreeRows  ' only exercises (left, middle, right side; no info line)
  ' the same with options - options never show any info line
  optLeft, optRight, optRightExt, optTwoRows, optTwoRowsExt, optThreeRows
end enum

declare sub arrayToList(l as TList, lpArray() as TListPoint)
declare function calculateLevel(addition as ulong, factor as single) as integer
declare function chooseWords(exercise as TExercise, byref settingString as string, _
                             opt as long = SETTING_OPTIONS.NONE) as long
declare function chooseSingleWords(exercise as TExercise, byref setting as string, _
                                   opt as long = SETTING_OPTIONS.NONE) as long
declare function chooseSingleWordsOutpt(x as integer, y as integer, x2 as integer, h as integer, _
                                        word as TWord, optional as integer) as integer
declare sub createMenu(id as string, byref gui as TGui, options as integer, _
                       cID1 as integer, sID1 as integer = -1, cID2 as integer = -1, sID2 as integer = -1)
declare sub exitModule
declare function getExercise(id as string) as TExercise
declare sub getInfo(byref module as TModule)
declare sub guiAddKeys(byref tempGui as TGui, sID as integer, keys() as string, sx as integer, sy as integer, _
                       w as integer, h as integer, dx as integer, dy as integer = 0, perLine as integer = 255)
declare sub guiAddLevs(byref tempGui as TGui, sID as integer, sx as integer, sy as integer, _
                       w as integer, h as integer, dx as integer, fsize as integer = FONTSIZE.X_SMALL)
declare function initModule(module as TModule ptr) as RETVAR
declare sub initDefaultLists(fullList as TList)
declare sub listToArray(l as TList, key as string, keyArray() as string, lpArray() as TListPoint)
declare sub mainWindow
declare function randomWord overload(words as TList = TList) as TWord
declare function randomWord(words as TList, byval count as integer) as mdList(TWord)
declare function searchInput(txt as string = "", x as integer = SCREEN_X\2 - 250, y as integer = SCREEN_Y\2-150, _
                             w as integer = 500, h as integer = 300) as string
declare function searchWord overload(word as string, pattern as string = "") as TWord
declare function searchWord(list as TList, word as string, pattern as string = "") as TWord
declare function standardInfo(exercise as TExercise) as string
declare sub wordToArray(txt as string, arr() as string, opt as integer = 1)

dim shared as mdList(TExercise)   globalExercises, globalOptions, globalHiddenExercises
dim shared as mdList(TWord) globallist_add
dim shared as mdMap(long, long)   listOfLessions
dim shared as mdMap(string, long) listOfTags

sub arrayToList(l as TList, lpArray() as TListPoint)
  l.list.clear
  for i as integer = 0 to ubound(lpArray)
    l.list.add(lpArray(i))
  next
end sub

function calculateLevel(addition as ulong, factor as single) as integer
  if addition < 1 or factor < 1.2 then return 0
  return 100-(1/99)^(.002*(addition\3+2)*factor - 1.005)
end function

function chooseWords(exercise as TExercise, byref settingString as string, _
                     opt as long = SETTING_OPTIONS.NONE) as long
  ' choose words for an exercise; return new settingString
  dim as integer lXstart =  20, lYstart =  80, lWidth =  60, lHeight = 22
  dim as integer lXplus  =  10, lYplus  =   5, lMod   =  11, lLines  = 10
  dim as integer tXstart =  20, tYstart = 380, tWidth = 140, tHeight = 22
  dim as integer tXplus  =  10, tYplus  =   5, tMod   =   5, tLines  =  8
  dim as integer mXstart = 140, mYStart =  20, mWidth =  40, mHeight = 22
  dim as integer mXPlus  =  20, mYPlus  =   0, mMod   =   8, mLines  =  1
  dim as integer mIndex = 50, fsize = FONTSIZE.TINY
  dim as integer vXstart =  20, vYstart = 430, vWidth = 119, vHeight = 25
  dim as integer vXplus  =  10, vYplus  =  10, vMod   =   6, vLines  =  3
  dim as integer vX1 = 150, vY1 = 340, vX2 = 280, vY2 = 340
  dim as integer okX = 650, okY = 600, okW = 120, okH = 25, okIndex = 99
  dim as integer chooseAllX = 30, chooseAllY = okY, chooseAllW = okW, chooseAllH = okH, chooseAllIndex = 98
  dim as integer chooseNoX = 200, chooseNoY  = okY, chooseNoW  = okW, chooseNoH  = okH, chooseNoIndex  = 97
  dim as integer helpX = 360, helpY = 590, helpIndex = 96
  dim as integer verbRegIndex = 95, verbIrregIndex = 94
  dim as TGuiElement tempButton
  dim as integer mx, my, mb, i
  dim as long lession, lCount = 0, tCount = 0
  dim as string key, tagname, tags(tMod*tLines-1)
  dim as mdList(long) ks1 = listOfLessions.keySet
  dim as TGui settingGui
  dim as TGuiEvent event

  if len(settingString) < lMod*lLines + tMod*tLines + 1 then
    settingString &= string((lMod*lLines + tMod*tLines + 1) - len(settingString), 0)
  end if
  if (opt and SETTING_OPTIONS.SKIPDIALOG) = 0 then
    cls
    printFT 20, 50, localize("chooseChapter"), font, FONTSIZE.SMALL
    if opt and SETTING_OPTIONS.SHOWTAGS then
      printFT 20, 340, localize("chooseGroups"), font, FONTSIZE.SMALL
    else
      if opt and SETTING_OPTIONS.SHOWVERBREG then
        printFT 20, 340, localize("verbs"), font, FONTSIZE.SMALL
      end if
    end if
    printFT 20, mYStart, localize("marks"), font, FONTSIZE.SMALL
  end if

  ' create GUI
  i = 0
  foreach(long, lession, ks1)
    if lession >= 100 then
      dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, lession, _
             lXstart+(lCount mod lMod)*(lWidth+lXplus), _
             lYStart+(lCount\lMod)*(lHeight+lYplus), lWidth, lHeight, format(lession/100, "###.00"))
      if settingString[i+1] > 0 then
        tempButton.state = GuiState.activated
      end if
      tempButton.fontsize = fsize
      settingGui.addElement tempButton
      lCount += 1
      if lCount = lMod*lLines then exit while
    end if
    i += 1
  nexteach
  i = 0
  dim as mdList(string) ks2 = listOfTags.keySet
  if opt and SETTING_OPTIONS.SHOWTAGS then
    foreach(string, tagname, ks2)
      if listOfTags.get(tagname) > 3 then
        tags(tCount) = tagname
        dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, tCount+1, _
               tXstart+(tCount mod tMod)*(tWidth+tXplus), _
               tYStart+(tCount\tMod)*(tHeight+tYplus), tWidth, tHeight, tagname)
        if settingString[lMod*lLines+tCount+1] > 0 then
          tempButton.state = GuiState.activated
        end if
        tempButton.fontsize = fsize
        settingGui.addElement tempButton
        tCount += 1
        if tCount = tMod*tLines then exit while
      end if
      i += 1
    nexteach
  else
    if opt and SETTING_OPTIONS.SHOWVERBREG then
      dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, verbRegIndex, _
                                      vX1, vY1, vWidth, vHeight, localize("reguarVerbs"))
      if settingString[lMod*lLines+1] and 1 then tempButton.state = GuiState.activated
      tempButton.fontsize = fsize
      settingGui.addElement tempButton
      tempButton = TGuiElement(GuiElementType.button, verbIrregIndex, vX2, vY2, vWidth, vHeight, localize("irreguarVerbs"))
      if settingString[lMod*lLines+1] and 2 then tempButton.state = GuiState.activated
      tempButton.fontsize = fsize
      settingGui.addElement tempButton
    end if
  end if
  i = 0
  for i as integer = 0 to 7
    tempButton = TGuiElement(GuiElementType.button, i+mIndex, mXstart+i*(mWidth+mXplus), mYStart, mWidth, mHeight, str(i+1))
    if bit(settingString[0], i) then tempButton.state = GuiState.activated
    tempButton.fontsize = fsize
    settingGui.addElement tempButton
  next
  tempButton = TGuiElement(GuiElementType.button, okIndex, okX, okY, okW, okH, localize("back"), chr(27))
  tempButton.fontsize = fsize
  settingGui.addElement tempButton
  tempButton = TGuiElement(GuiElementType.button, chooseNoIndex, _
               chooseNoX,  chooseNoY,  chooseNoW,  chooseNoH,  localize("chooseNone"))
  tempButton.fontsize = fsize
  settingGui.addElement(tempButton)
  tempButton = TGuiElement(GuiElementType.button, chooseAllIndex, _
               chooseAllX, chooseAllY, chooseAllW, chooseAllH, localize("chooseAllChapters"))
  tempButton.fontsize = fsize
  settingGui.addElement(tempButton)
  settingGui.addElement TGuiElement(GuiElementType.button, helpIndex, _
                        exepath & "/img/help.bmp", helpX, helpY, 40, 40, chr(255, 59),,, GuiState.noBorder)
  ' check input
  if (opt and SETTING_OPTIONS.SKIPDIALOG) = 0 then
    settingGui.show
    do
      getmouse mx, my,, mb
      sleep 1
    loop until mb < 1
    do
      key = inkey
      event = settingGui.getEvent(key)
      if event.mbutton = 1 orelse len(event.key) then
        select case event.id
          case okIndex
          case chooseNoIndex
            ' toggle off marks
            for i as integer = 0 to 7
              tempButton = TGuiElement(GuiElementType.button, i+mIndex, _
                           mXstart+i*(mWidth+mXplus), mYStart, mWidth, mHeight, str(i+1))
              tempButton.state = GuiState.active
              tempButton.fontsize = fsize
              settingGui.setElement tempButton
            next
            ' toggle off lessions
            dim as integer i = 0
            foreach(long, lession, ks1)
              if lession >= 100 then
                dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, lession, _
                       lXstart+(i mod lMod)*(lWidth+lXplus), _
                       lYStart+(i\lMod)*(lHeight+lYplus), lWidth, lHeight, format(lession/100, "###.00"))
                tempButton.state = GuiState.active
                tempButton.fontsize = fsize
                settingGui.setElement tempButton
                i += 1
                if i >= lCount then exit while
              end if
            nexteach
            i = 0
            ' toggle off tags
            if opt and SETTING_OPTIONS.SHOWTAGS then
              foreach(string, tagname, ks2)
                if listOfTags.get(tagname) > 3 then
                  tags(tCount) = tagname
                  dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, i+1, _
                         tXstart+(i mod tMod)*(tWidth+tXplus), _
                         tYStart+(i\tMod)*(tHeight+tYplus), tWidth, tHeight, tagname)
                  tempButton.state = GuiState.active
                  tempButton.fontsize = fsize
                  settingGui.setElement tempButton
                  i += 1
                  if i >= lCount then exit while
                end if
              nexteach
            end if
            ' toggle off verb settings
            if opt and SETTING_OPTIONS.SHOWVERBREG then
              dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, verbRegIndex, _
                     vX1, vY1, vWidth, vHeight, localize("reguarVerbs"))
              tempButton.fontsize = fsize
              settingGui.setElement tempButton
              tempButton = TGuiElement(GuiElementType.button, verbIrregIndex, _
                           vX2, vY2, vWidth, vHeight, localize("irreguarVerbs"))
              tempButton.fontsize = fsize
              settingGui.setElement tempButton
            end if
            settingGui.show
          case chooseAllIndex
            ' toggle on lessions
            dim as integer i = 0
            foreach(long, lession, ks1)
              if lession >= 100 then
                dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, lession, _
                       lXstart+(i mod lMod)*(lWidth+lXplus), _
                       lYStart+(i\lMod)*(lHeight+lYplus), lWidth, lHeight, format(lession/100, "###.00"))
                tempButton.state = GuiState.activated
                tempButton.fontsize = fsize
                settingGui.setElement tempButton
                i += 1
                if i >= lCount then exit while
              end if
            nexteach
            settingGui.show
          case helpIndex
            dim as integer ret = showHelpScreen(MODULE_PATH & "doc.htm", "optionsDefault")
            if ret = RETVAR.windowClosed then key = chr(255, 107) : exit do
          case else
            dim as TGuiElement tempButton = settingGui.getElement(event.id)
            tempButton.state xor= GuiState.activated
            settingGui.setElement(tempButton)
            tempButton.show
        end select
      end if
      do
        getmouse mx, my,, mb
        sleep 1
      loop until mb < 1
      if event.id = okIndex and (key = chr(27) or checkMouseIn(okX, okY, okW, okH) <> 0) then exit do
    loop until key = chr(255, 107)
  end if

  ' create exercise.currentList
  exercise.currentList.list.clear
  settingString = string((lMod*lLines + tMod*tLines + 1), 0)
  dim lessions as mdList(long), countKey as integer = 0
  dim as integer useRegulars = 0, useIrregulars = 0
  redim as string keys(ubound(tags)+8), values(ubound(tags)+8)
  foreach(TGuiElement, tempButton, settingGui.elements)
    if tempButton.state and GuiState.activated then
      select case tempButton.id
        case 1 to tCount
          keys(countKey) = "tag"
          values(countKey) = tags(tempButton.id-1)
          countKey += 1
          settingString[lMod*lLines+tempButton.id] = 1
        case mIndex to mIndex+7
          keys(countKey) = "mark"
          values(countKey) = str(tempbutton.id-mIndex+1)
          countKey += 1
          settingString[0] = bitset(settingString[0], tempButton.id-mIndex)
        case verbRegIndex
          useRegulars = 1
        case verbIrregIndex
          useIrregulars = 2
        case is >= 10000
          settingString[lMod*lLines+tempButton.id-9998] = 1
        case is >= 100
          lessions.add(tempButton.id)
          dim as integer i = 0
          foreach(long, lession, ks1)
            if lession = tempButton.id then exit while
            i += 1
          nexteach
          if i < ks1.size then settingString[i+1] = 1
      end select
    end if
  nexteach
  if countKey > 0 then redim preserve keys(countKey-1) else erase keys
  if lessions.size = 0 andalso countKey = 0 then
    exercise.currentList.list.addAll(exercise.fullList.list)
  else
    exercise.currentList.add, lessions, keys(), values(), Operators.or_
  end if
  if SETTING_OPTIONS.SHOWVERBREG then
    settingString[lMod*lLines+1] = useRegulars or useIrregulars
    select case useRegulars or useIrregulars
      case 1 : exercise.currentList.del 0, &hffffff, "irr", ""
      case 2 : exercise.currentList.remain "irr", ""
    end select
  end if
  exercise.updateLists
  if key = chr(255, 107) then return RETVAR.windowClosed
  return RETVAR.ok
end function

function chooseSingleWords(exercise as TExercise, byref setting as string, _
                           opt as long = SETTING_OPTIONS.NONE) as long
  enum
    posUp = 1, posDown, posAll, posNone, posHelp, posBack
  end enum
  dim as integer LBx = 100, LBy = 590, LBw = 110, LBh = 30, LBp = 20, upState, downState, mx, my, mb, mw, wheel
  dim as string key

  dim as mdList(long) marked
  for i as integer = 0 to len(setting)-3 step 3
    marked.add (setting[i] shl 16) + (setting[i+1] shl 8) + setting[i+2]
  next

  if (opt and SETTING_OPTIONS.SKIPDIALOG) = 0 then

    dim chooseGui as TGui, event as TGuiEvent
    dim as mdList(TListPoint) list
    chooseGui.addElement _
        TGuiElement(GuiElementType.button, posAll, LBx+(LBw+LBp), LBy, LBw, LBh, localize("chooseAll"))
    chooseGui.addElement _
        TGuiElement(GuiElementType.button, posNone, LBx+2*(LBw+LBp), LBy, LBw, LBh, localize("chooseNone"))
    chooseGui.addElement _
        TGuiElement(GuiElementType.button, posBack, LBx+4*(LBw+LBp), LBy, LBw, LBh, localize("back"), chr(27))
    chooseGui.addElement _
        TGuiElement(GuiElementType.button, posHelp, _
        exepath & "/img/help.bmp", LBx+50, LBy, 40, 40, chr(255, 59),, "", GuiState.noBorder)
    chooseGui.addElement _
        TGuiElement(GuiElementType.button, posUp, _
                    exepath & "/img/up.bmp", LBx-60, LBy, 31, 30, chr(255, 73))
    chooseGui.addElement _
        TGuiElement(GuiElementType.button, posDown, _
                    exepath & "/img/down.bmp", LBx-20, LBy, 31, 30, chr(255, 81))
    chooseGui.addElement _
        TGuiElement(GuiElementType.button, posDown, _
                    exepath & "/img/down.bmp", LBx-20, LBy, 31, 30, chr(255, 81))
    list.addAll(exercise.fullList.list)
    if list.size = 0 then return RETVAR.emptyList
    ' show words
    dim as TPages pages = TPages(13, 70, 15, 15, 10, 60, 550, 190, 4)
    pages.outp = @chooseSingleWordsOutpt
    pages.calcPositions(list)
    dim as integer page = 0, lastMarked = -1, markDir = -1
    dim as FB.Image ptr okImg = imagecreate(16, 16), emptyImg = imagecreate(16, 16, rgb(255, 255, 255))
    bload exepath & "/img/ok.bmp", okImg

    do
      screenlock
      cls
      line (0, 0)-step(SCREEN_X-1, LBy-1), &hffffff, bf
      pages.printEntries(list, page)
      for i as integer = 0 to ubound(pages.wordPos)
        with pages.wordPos(i)
          if marked.contains(.id) then put (.x-30, .y), okImg, pset
        end with
      next
      
      ' show arrow buttons or not?
      if pages.pages.size = 1 then
        ' hide if only one page is shown
        if upState <> GuiState.hidden then
          dim as TGuiElement tempButton = chooseGui.getElement(posUp)
          tempButton.state = GuiState.hidden
          chooseGui.setElement(tempButton)
          upstate = GuiState.hidden
        end if
        if downState <> GuiState.hidden then
          dim as TGuiElement tempButton = chooseGui.getElement(posDown)
          tempButton.state = GuiState.hidden
          chooseGui.setElement(tempButton)
          downState = GuiState.hidden
        end if
      else
        if page = 0 then
          ' deactivate "up"
          if upState <> GuiState.deactivated then
            dim as TGuiElement tempButton = chooseGui.getElement(posUp)
            tempButton.state = GuiState.deactivated
            chooseGui.setElement(tempButton)
            upState = GuiState.deactivated
          end if
        else
          ' activate "up"
          if upState <> GuiState.active then
            dim as TGuiElement tempButton = chooseGui.getElement(posUp)
            tempButton.state = GuiState.active
            chooseGui.setElement(tempButton)
            upState = GuiState.active
          end if
        end if
        if page = pages.pages.size-1 then
          ' deactivate "down"
          if downState <> GuiState.deactivated then
            dim as TGuiElement tempButton = chooseGui.getElement(posDown)
            tempButton.state = GuiState.deactivated
            chooseGui.setElement(tempButton)
            downState = GuiState.deactivated
          end if
        else
          ' activate "down"
          if downState <> GuiState.active then
            dim as TGuiElement tempButton = chooseGui.getElement(posDown)
            tempButton.state = GuiState.active
            chooseGui.setElement(tempButton)
            downState = GuiState.active
          end if
        end if
      end if
      chooseGui.show
      screenunlock
      do
        getmouse mx, my, wheel, mb
        sleep 1
      loop until mb < 1
      ' user input
      do
        key = inkey
        event = chooseGui.getEvent(key)
        ' mouse wheel / cursor
        if getmouse(mx, my, mw, mb) then mb = 0 : mw = wheel
        if mw > wheel or key = chr(255, 72) then
          wheel = mw
          event.id = posUp
        elseif mw < wheel or key = chr(255, 80) then
          wheel = mw
          event.id = posDown
        end if
        ' mark words
        if mb = 0 then lastMarked = -1 : markDir = -1
        if mb = 1 then
          for i as integer = 0 to ubound(pages.wordPos)
            with pages.wordPos(i)
              if checkMouseIn(.x, .y-pages.wordDist\2, pages.colDist\2, .h+pages.wordDist) <> 0 then
                if lastMarked <> i then
                  if markDir <> 1 andalso marked.contains(.id) = -1 then
                    markDir = 0
                    marked.remove .id
                    put (.x-30, .y), emptyImg, pset
                  elseif markDir <> 0 andalso marked.contains(.id) = 0 then
                    markDir = 1
                    marked.add .id
                    put (.x-30, .y), okImg, pset
                  end if
                end if
                lastMarked = i
                exit for
              end if
            end with
          next
        end if
        ' GUI buttons
        select case event.id
          case posUp
            if page > 0 then
              page -= 1
              exit do
            end if
          case posDown
            if page < pages.pages.size-1 then
              page += 1
              exit do
            end if
          case posHelp
            dim as integer ret = showHelpScreen(MODULE_PATH & "doc.htm", "optionsSingleWords")
            if ret = RETVAR.windowClosed then key = chr(255, 107)
          case posAll
            marked.clear
            for i as integer = 0 to exercise.fullList.list.size - 1
              marked.add exercise.fullList.list.get(i).id
            next
            exit do
          case posNone
            marked.clear
            exit do
          case posBack
            exit do, do
        end select
        sleep 1
      loop until key = chr(255, 107)
    loop until key = chr(27) or key = chr(255, 107)
    imagedestroy okImg
    imagedestroy emptyImg

  end if

  dim as long i
  exercise.currentList.list.clear
  setting = ""
  if marked.isEmpty then
    exercise.currentList.list.addAll exercise.fullList.list
  else
    foreach(long, i, marked)
      exercise.currentList.list.add i
      setting &= chr(i shr 16) & chr((i shr 8) and 255) & chr(i and 255)
    nexteach
  end if
  exercise.currentList.size = exercise.currentList.list.size
  marked.clear
  if key = chr(255, 107) then return RETVAR.windowClosed
  return RETVAR.ok
end function

function chooseSingleWordsOutpt(x as integer, y as integer, x2 as integer, h as integer, _
                                word as TWord, optional as integer) as integer
  dim as integer col
  dim as string tempString, irr = word.map.get("irr"), w = word.map.get("root")
  if len(w) = 0 then w = word.word
  if optional mod 32 then
    printFT x, y, w, font, optional mod 32, col
    if len(irr) then
      dim as FB.Image ptr attImg = imagecreate(16, 16)
      bload exepath & "/img/attention.bmp", attImg
      put (x-14, y), attImg, pset
      imagedestroy attImg
    end if
  end if
  return h
end function

function createCategory(exercise as TExercise) as long
  dim as integer mx, my, mb
  dim as string key, settingID, settingString
  dim as TGuiEvent event

  dim as TGui gui
  createMenu exercise.id, gui, iif(exercise.opt = 0, MENUTYPE.exercThreeRows, exercise.opt), 1, 101

  gui.addElement TGuiElement(GuiElementType.button, 999, SCREEN_X-250, SCREEN_Y-60, 150, 30, localize("back"), chr(27))
  screenlock
  cls
  centerFT SCREEN_X\2, 50, exercise.headline, font, 20
  gui.show
  screenunlock
  
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  do
    key = inkey
    if key = chr(255, 107) then exit do
    event = gui.getEvent(key)
    select case event.id
      case 1 to ubound(globalExercArray)+1
        if globalExercArray(event.id-1).callback(globalExercArray(event.id-1)) = RETVAR.windowClosed then
          return RETVAR.windowClosed
        end if
        screenlock
        cls
        centerFT SCREEN_X\2, 50, exercise.title, font, 20
        gui.show
        screenunlock
        do
          getmouse mx, my,, mb
          sleep 1
        loop until mb = 0
      case 101 to ubound(globalExercArray)+101
        if globalExercArray(event.id-101).settings(globalExercArray(event.id-101), SETTING_OPTIONS.NONE) = RETVAR.windowClosed then
          return RETVAR.windowClosed
        end if
        screenlock
        cls
        centerFT SCREEN_X\2, 50, exercise.title, font, 20
        gui.show
        screenunlock
        do
          getmouse mx, my,, mb
          sleep 1
        loop until mb = 0
      case 999
        exit do
    end select
    sleep 1
  loop

  if key = chr(255, 107) then return RETVAR.windowClosed
  return RETVAR.ok
end function

sub createMenu(id as string, byref gui as TGui, options as integer, _
               cID1 as integer, sID1 as integer = -1, cID2 as integer = -1, sID2 as integer = -1)
  ' create menu buttons from globalExercArray and/or globalOptArray
  dim as integer b_startTop = 100, b_width = 150, b_height = 30, b_dist = 20
  dim as integer b_startLeft = 50, b_startRight = SCREEN_X-(b_startLeft+b_width+b_height+b_dist)
  dim as integer b_startRightExt = (SCREEN_X + b_startLeft)\2, b_startMiddle = (SCREEN_X-b_width-b_height-b_dist)\2
  dim as integer maxEntries = (SCREEN_Y - 2*b_startTop)\(b_height + b_dist)
  dim as integer b_endBottom = b_startTop + maxEntries*(b_height + b_dist)
  dim as integer exercX = -1, optX = -1, posX, posY, count
  select case options
    case MENUTYPE.both
      exercX = b_startLeft
      optX   = b_startRight
    case MENUTYPE.exercLeft, exercTwoRows, exercTwoRowsExt, exercThreeRows
      exercX = b_startLeft
    case MENUTYPE.exercRightExt
      exercX = b_startRightExt
    case MENUTYPE.exercRight
      exercX = b_startRight
    case MENUTYPE.optLeft, optTwoRows, optTwoRowsExt, optThreeRows
      optX   = b_startLeft
    case MENUTYPE.optRight
      optX   = b_startRight
    case MENUTYPE.optRightExt
      optX = b_startRightExt
  end select

  if exercX > 0 then
    count = 0
    posX = exercX
    posY = b_startTop
    for i as integer = 0 to ubound(globalExercArray)
      if globalExercArray(i).menu = id then
        gui.addElement(TGuiElement(GuiElementType.button, i+cID1, posX, posY, b_width, b_height, globalExercArray(i).title))
        if sID1 >= 0 andalso globalExercArray(i).settings then
          gui.addElement(TGuiElement(GuiElementType.button, i+sID1, _
              exepath & "/img/settings.bmp", posX + b_width + b_dist, posY, b_height, b_height))
        end if
        if sID1 >= 0 andalso globalExercArray(i).info then
          gui.addElement(TGuiElement(GuiElementType.label, -1, posX+b_width+b_dist+40, posY+b_height\2-6,,, _
                         @globalExercArray(i),,,,, FONTSIZE.TINY))
        end if
        count += 1
        posY += b_height+b_dist
        if posY > b_endBottom then
          select case options
            case MENUTYPE.both, MENUTYPE.exercLeft, MENUTYPE.exercRight, MENUTYPE.exercRightExt
              exit for
            case MENUTYPE.exercTwoRows
              if posX = b_startLeft then
                posY = b_startTop
                posX = b_startRight
              else
                exit for
              end if
            case MENUTYPE.exercTwoRowsExt
              if posX = b_startLeft then
                posY = b_startTop
                posX = b_startRightExt
              else
                exit for
              end if
            case MENUTYPE.exercThreeRows
              if posX = b_startLeft then
                posY = b_startTop
                posX = b_startMiddle
              elseif posX = b_startMiddle then
                posY = b_startTop
                posX = b_startRight
              else
                exit for
              end if
          end select
        end if
      end if
    next
  end if
  if optX > 0 then
    count = 0
    posX = optX
    posY = b_startTop
    for i as integer = 0 to ubound(globalOptArray)
      if globalOptArray(i).menu = id then
        gui.addElement(TGuiElement(GuiElementType.button, i+iif(exercX > 0, cID2, cID1), _
                        posX, posY, b_width, b_height, globalOptArray(i).title))
        if iif(exercX > 0, sID2, sID1) >= 0 andalso globalOptArray(i).settings then
          gui.addElement(TGuiElement(GuiElementType.button, i+iif(exercX > 0, sID2, sID1), exepath & "/img/settings.bmp", _
              posX + b_width + b_dist, posY, b_height, b_height))
        end if
        count += 1
        posY += b_height+b_dist
        if posY > b_endBottom then
          select case options
            case MENUTYPE.both, MENUTYPE.optLeft, MENUTYPE.optRight, MENUTYPE.optRightExt
              exit for
            case MENUTYPE.optTwoRows
              if posX = b_startLeft then
                posY = b_startTop
                posX = b_startRight
              else
                exit for
              end if
            case MENUTYPE.optTwoRowsExt
              if posX = b_startLeft then
                posY = b_startTop
                posX = b_startRightExt
              else
                exit for
              end if
            case MENUTYPE.optThreeRows
              if posX = b_startLeft then
                posY = b_startTop
                posX = b_startMiddle
              elseif posX = b_startMiddle then
                posY = b_startTop
                posX = b_startRight
              else
                exit for
              end if
          end select
        end if
      end if
    next
  end if
end sub

sub exitModule
  ' END OF PROGRAM: SAVE CHANGES
  kill MODULE_PATH & "words.bak"
  name MODULE_PATH & "words.txt",  MODULE_PATH & "words.bak"
  dim as integer filenumber = freefile
  dim as long tempLong
  open MODULE_PATH & "words.txt" for output as #filenumber
  if len(WORDS_VERSION) then print #filenumber, "# VERSION "; WORDS_VERSION; chr(13, 10); chr(13, 10);
  dim as mdSet(long) ks = globallist.keyset
  foreach(long, tempLong, ks)
    print #filenumber, str(globallist.get(tempLong)); chr(13, 10);
  nexteach
  close #filenumber
  screen 0
end sub

function getExercise(id as string) as TExercise
  ' return the exercise (or option) with the given id
  for i as integer = 0 to ubound(globalExercArray)
    if globalExercArray(i).id = id then return globalExercArray(i)
  next
  for i as integer = 0 to ubound(globalOptArray)
    if globalOptArray(i).id = id then return globalOptArray(i)
  next
  return TExercise("", "", "full", 0, 0, 0, 0)
end function

sub getInfo(byref module as TModule) export
  module.libShort = LANGUAGE
  module.title    = LANGTITLE1
  module.major    = moduleVersionMajor
  module.minor    = moduleVersionMinor
  module.patch    = moduleVersionPatch
  module.api      = API_VERSION
end sub

sub guiAddKeys(byref tempGui as TGui, sID as integer, keys() as string, sx as integer, sy as integer, _
               w as integer, h as integer, dx as integer, dy as integer = 0, perLine as integer = 255)
  ' add buttons for special keys to a GUI
  ' keys() : array of special keys (UTF8)
  ' sx/sy  : start of the first button
  ' w /h   : widht/height of each button
  ' dx/dy  : distance between the buttons (horizontal/vertical)
  ' perLine: number of buttons per line (number of lines will be calculated automatically)
  for i as integer = lbound(keys) to ubound(keys)
    if len(keys(i)) andalso keys(i) <> " " then
      tempGui.addElement TGuiElement(GuiElementType.button, sID+i, _
                         sx+(w+dx)*(i mod perLine), sy+(h+dy)*(i\perLine), w, h, keys(i))
    end if
  next
end sub

sub guiAddLevs(byref tempGui as TGui, sID as integer, sx as integer, sy as integer, _
               w as integer, h as integer, dx as integer, fsize as integer = FONTSIZE.X_SMALL)
  ' add buttons to rate the difficulty of a word
  dim as string txt(...) = { localize("lvl1"), localize("lvl2"), localize("lvl3"), _
                             localize("lvl4"), localize("lvl5"), localize("lvl6") }
  for i as integer = 0 to ubound(txt)
    dim as TGuiElement tempButton = TGuiElement(GuiElementType.button, _
                                    sID+i, sx+(w+dx)*i, sy, w, h, txt(i), chr(255, 62+i))
    tempButton.fontsize = fsize
    tempButton.info = "F" & i+4
    tempGui.addElement tempButton
  next
end sub

function initModule(module as TModule ptr) as RETVAR
  ' initiate module
  screenres SCREEN_X, SCREEN_Y, 32
  width SCREEN_X\8, SCREEN_Y\8
  color 0, &hffffff
  randomize
  windowtitle module->title & " " & LANGTITLE1 & " - " & LANGTITLE2 _
              & " (v" & moduleVersionMajor & "." & moduleVersionMinor & "." & moduleVersionPatch & ")"
  globallist = mdMap(long, TWord)

  ' INITIALIZE FONT
  FT_Var.ErrorMsg = FT_Init_FreeType(@FT_Var.Library)
  if FT_Var.ErrorMsg then print localize("noFontLib") : getkey : return RETVAR.missingLib
  font = getFont(exepath & "/font.ttf")
  if font = 0 then print localize("noFontFound") : getkey : return RETVAR.missingFont

  cls
  printFT 100, SCREEN_Y\2-50, localize("loadData"), font
  drawProgress 100, SCREEN_Y\2-10, SCREEN_X-200, 20, 0, 1

  dim as integer filenumber = freefile, search
  dim tempString as string, tempWord as TWord, tempLong as long

  ' READ OPTIONS
  if open(OPTIONS_FILENAME for input as #filenumber) = 0 then
    do until eof(filenumber)
      line input #fileNumber, tempString
      select case tempString
        case "EDITS=ALLOWED" : GlobalSettings or= GLOBAL_SETTINGS.ALLOW_EDITS
      end select
    loop
    close #filenumber
  end if

  ' READ WORDS
  open MODULE_PATH & "words.txt" for input as #filenumber
  do until eof(filenumber)
    line input #filenumber, tempString
    if left(lcase(tempString), 10) = "# version " then
      WORDS_VERSION = trim(mid(tempString, 11))
      line input #filenumber, tempString
    end if
    select case lcase(left(tempString, 6))
      case "entry "
        if tempWord.id then globallist.put(tempWord.id, tempWord)
        tempWord.clear
        tempWord.id = cint(mid(tempString, 7))
        if globallist.containsKey(tempWord.id) then print "error: double entry " & tempWord.id
        tempWord.lession = tempWord.id \ 1000
        tempWord.lessionPos = tempWord.id mod 1000
        if listOfLessions.containsKey(tempWord.lession) then
          listOfLessions[tempWord.lession] += 1
        else
          listOfLessions.put(tempWord.lession, 1)
        end if
      case "word  "
        tempWord.word = trim(mid(tempString, 7))
      case "mark  "
        tempWord.mark = cint(mid(tempString, 7))
      case "group "
        tempWord.groupname = trim(mid(tempString, 7))
      case "trans "
        tempWord.translate.add(trim(mid(tempString, 7)))
      case "pos   "
        tempWord.lession = cint(mid(tempString, 7)) \ 1000
        tempWord.lessionPos = cint(mid(tempString, 7)) mod 1000
      case else
        if lcase(left(tempString, 4)) = "tag " then
          tempString = trim(mid(tempString, 5))
          tempWord.tag.add(tempString)
          if listOfTags.containsKey(tempString) then
            listOfTags[tempString] += 1
          else
            listOfTags.put(tempString, 1)
          end if
        else
          search = instr(tempString, " ")
          tempWord.map.put(left(tempString, search-1), trim(mid(tempString, search+1)))
        end if
    end select
  loop
  if tempWord.id then globallist.put(tempWord.id, tempWord)
  tempWord.clear
  close #filenumber
  
  ' READ ADDITIONAL WORDLIST (<LANGSHORT2>/words.txt)
  dim as string value
  if open(exepath & "/modules/add_" & LANGSHORT2 & "/words.txt" for input as #filenumber) = 0 then
    do until eof(filenumber)
      line input #filenumber, tempString
      if left(tempString, 1) = "#" then continue do
      if left(tempString, 6) = "word  " then
        if len(tempWord.word) then globallist_add.add tempWord
        tempWord.clear
        tempWord.word = mid(tempString, 7)
      else
        search = instr(tempString, " ")
        value  = mid(tempString, search+1)
        select case left(tempString, search-1)
          case "group"
            tempWord.groupname = value
          case "trans"
            tempWord.translate.add(value)
          case "tag"
            tempWord.tag.add(value)
          case else
            tempWord.map.put(left(tempString, search-1), value)
        end select
      end if
    loop
    if len(tempWord.word) then globallist_add.add tempWord
    tempWord.clear
    close #filenumber
  end if

  return RETVAR.ok
end function

sub initDefaultLists(fullList as TList)
  ' update lists
  dim as TList list(1)
  dim as TListPoint listPoint
  dim as integer sum = 0
  for i as integer = 0 to 1
    with list(i)
      .listName = "default-" & iif(i = 0, LANGSHORT1 & "-" & LANGSHORT2, LANGSHORT2 & "-" & LANGSHORT1)
      .load
      ' add new words
      foreach(TListPoint, listPoint, fullList.list)
        if .list.contains(listPoint) = 0 then
          listPoint.position = 0
          if .list.isEmpty then
            .list.add(listPoint)
          else
            .list.add(0, listPoint)
          end if
        end if
      nexteach
      ' delete words that no longer exist
      foreach(TListPoint, listPoint, .list)
        if fullList.list.contains(listPoint) = 0 then
          .list.remove(listPoint)
        end if
      nexteach
      .size = .list.size
      .sort
      .mix
      .save
    end with
  next
  for i as integer = 0 to 1
    with list(i)
      dim as integer k = 0
      foreach(TListPoint, listPoint, .list)
        dim as TWord tempWord = globallist.get(listPoint.id)
        if tempWord.id then
          if i = 0 then
            tempWord.levelA = calculateLevel(listPoint.addition, listPoint.factor)
            tempWord.countA = listPoint.special
          else
            tempWord.levelB = calculateLevel(listPoint.addition, listPoint.factor)
            tempWord.countB = listPoint.special
          end if
          globallist.put(listPoint.id, tempWord)
        else
          fullList.list.remove(listPoint)
        end if
        k += 1
        sum = iif(sum >= .size*2, .size*2, sum+1)
        screenlock
        drawProgress 100, SCREEN_Y\2-10, SCREEN_X-200, 20, sum, .size*2
        line (0, SCREEN_Y\2+40)-step(SCREEN_X-1, 50), rgb(255, 255, 255), bf
        centerFT SCREEN_X\2, SCREEN_Y\2+50, localize("XofYwords", str(sum\2), str(.size)), font
        screenunlock
      nexteach
      .list.clear
    end with
  next
  fullList.save
end sub

sub listToArray(l as TList, key as string, keyArray() as string, lpArray() as TListPoint)
  ' produces two arrays for quicksort method
  ' l:        list of words you want to prepare for quicksort
  ' key:      the attribute you will use for sorting
  ' keyArray: array which is ready for sorting
  ' lpArray:  array of listpoints corresponding to keyArray
  '           In the next step you can quicksort keyArray and lpArray together and then easily re-produce a sorted list.
  dim as TListPoint listPoint
  dim as TWord w
  dim as integer i = 0
  redim keyArray(l.list.size-1), lpArray(l.list.size-1)
  foreach(TListPoint, listPoint, l.list)
    w = globallist.get(listPoint.id)
    select case key
      case "word"
        keyArray(i) = modifyTxt(w.word, MODIFY_SORTABLE)
      case "mark"
        if w.mark then
          keyArray(i) = chr(w.mark)
        else
          keyArray(i) = chr(255)
        end if
      case "trans"
        keyArray(i) = modifyTxt(w.translate.get(0), MODIFY_SORTABLE)
      case "group"
        keyArray(i) = modifyTxt(w.groupname, MODIFY_SORTABLE)
      case "tag"
        keyArray(i) = modifyTxt(w.tag.get(0), MODIFY_SORTABLE)
      case "levelA"
        keyArray(i) = chr(w.levelA)
      case "levelB"
        keyArray(i) = chr(w.levelB)
      case "id"
        keyArray(i) = format(w.lession*1000 + w.lessionPos, "00000000000")
      case else
        keyArray(i) = modifyTxt(w.map.get(key))
    end select
    lpArray(i) = listPoint
    i += 1
  nexteach
end sub

sub mainWindow
  ' the main program window: start an exercise, call the settings and so on
  ' choose and execute a test
  dim key as string, event as TGuiEvent, ret as integer
  ' buttons
  dim as integer mb_count = globalExercises.size, mo_count = globalOptions.size
  dim as integer mb_startX = 50, mb_startY = 100, mb_width = 150, mb_height = 30, mb_dist = 20
  dim as integer mo_startX = 550, mo_startY = 100, mo_width = 150, mo_height = 30, mo_dist = 20
  redim globalExercArray(mb_count-1), globalOptArray(mo_count-1)
  dim as TList spec
  dim as long tempLong, filenumber = freefile
  dim as TGui mainGui
  dim as integer exIndex = 100, optIndex = 150, setIndex = 100, infoIndex = 901, exitIndex = 902, mx, my, mb
  dim sizeID as long, sizeData as longint, stringID as string, stringData as string

  ' initialize Exercises
  centerFT SCREEN_X\2, SCREEN_Y\2+80, localize("initExerc"), font
  for i as integer = 0 to mb_count-1
    drawProgress 100, SCREEN_Y\2-10, SCREEN_X-200, 20, i+1, (mb_count+mo_count)*2
    globalExercArray(i) = globalExercises.get(i)
    globalExercArray(i).fullList = TList(globalExercArray(i).listname)
    globalExercArray(i).fullList.load
  next
  ' initialize options
  for i as integer = 0 to mo_count-1
    drawProgress 100, SCREEN_Y\2-10, SCREEN_X-200, 20, mb_count+i+1, (mb_count+mo_count)*2
    globalOptArray(i) = globalOptions.get(i)
    globalOptArray(i).fullList = TList(globalOptArray(i).listname)
    globalOptArray(i).fullList.load
  next
  createMenu "main", mainGui, MENUTYPE.both, exIndex, exIndex+setIndex, optIndex, optIndex+setIndex

  ' read setting strings
  open MODULE_PATH & "exercise.dat" for binary as #filenumber
  do until eof(filenumber)
    get #filenumber,, sizeID
    if lof(filenumber)-loc(filenumber) < sizeID then exit do
    if sizeID > 0 then
      stringID = space(sizeID)
      get #filenumber,, stringID
    end if
    if eof(filenumber) then exit do
    get #filenumber,, sizeData
    if lof(filenumber)-loc(filenumber) < sizeData then exit do
    if sizeData > 0 then
      stringData = space(sizeData)
      get #filenumber,, stringData
    end if
    for i as integer = 0 to mb_count-1
      drawProgress 100, SCREEN_Y\2-10, SCREEN_X-200, 20, mb_count+mo_count+i+1, (mb_count+mo_count)*2
      if globalExercArray(i).id = stringID then
        globalExercArray(i).settingString = stringData
        if globalExercArray(i).settings then
          globalExercArray(i).settings(globalExercArray(i), SETTING_OPTIONS.SKIPDIALOG)
        end if
        globalExercArray(i).currentList.update
      end if
    next
    for i as integer = 0 to mo_count-1
      drawProgress 100, SCREEN_Y\2-10, SCREEN_X-200, 20, mb_count*2+mo_count+i+1, (mb_count+mo_count)*2
      if globalOptArray(i).id = stringID then
        globalOptArray(i).settingString = stringData
        if globalOptArray(i).settings then 
          globalOptArray(i).settings(globalOptArray(i), SETTING_OPTIONS.SKIPDIALOG)
        end if
        globalOptArray(i).currentList.update
      end if
    next
  loop
  close filenumber

  mainGui.addElement TGuiElement(GuiElementType.button, infoIndex, _
                     mo_startX, 500, mo_width, mo_height, localize("about"), chr(255, 59))
  mainGui.addElement TGuiElement(GuiElementType.button, exitIndex, _
                     mo_startX, 550, mo_width, mo_height, localize("end"))
  do
    ' draw GUI
    screenlock
    cls
    centerFT 400, 50, localize("mainHeadline", LANGTITLE1, LANGTITLE2), font, FONTSIZE.HEADLINE
    mainGui.show
    screenunlock
    ' check input
    do
      getmouse mx, my,, mb
      sleep 1
    loop until mb < 1
    do
      key = inkey
      event = mainGui.getEvent(key)
      select case event.id
        case exIndex to exIndex+mb_count-1
          ret = globalExercArray(event.id-exIndex).callback(globalExercArray(event.id-exIndex))
          globalExercArray(event.id-exIndex).fullList.lastSave = now
          globalExercArray(event.id-exIndex).fullList.save
          exit do
        case exIndex+setIndex to exIndex+setIndex+mb_count-1
          ret = globalExercArray(event.id-exIndex-setIndex).settings _
                (globalExercArray(event.id-exIndex-setIndex), SETTING_OPTIONS.NONE)
          exit do
        case optIndex to optIndex+mo_count-1
          ret = globalOptArray(event.id-optIndex).callback(globalOptArray(event.id-optIndex))
          exit do
        case optIndex+setIndex to optIndex+setIndex+mo_count-1
          ret = globalOptArray(event.id-optIndex-setIndex).settings _
                (globalOptArray(event.id-optIndex-setIndex), SETTING_OPTIONS.NONE)
          exit do
        case infoIndex
          ret = showHelpScreen(MODULE_PATH & "doc.htm", "about")
          exit do
        case exitIndex
          ret = RETVAR.windowClosed
      end select
      sleep 1
    loop until key = chr(255, 107) or ret = RETVAR.windowClosed
  loop until key = chr(255, 107) or ret = RETVAR.windowClosed
  ' save setting strings
  kill MODULE_PATH & "exercise.dat"
  open MODULE_PATH & "exercise.dat" for binary as #filenumber
  for i as integer = 0 to mb_count-1
    put #filenumber,, cast(long, len(globalExercArray(i).id))
    put #filenumber,, globalExercArray(i).id
    put #filenumber,, cast(longint, len(globalExercArray(i).settingString))
    put #filenumber,, globalExercArray(i).settingString
    globalExercArray(i).fullList.save
  next
  for i as integer = 0 to mo_count-1
    put #filenumber,, cast(long, len(globalOptArray(i).id))
    put #filenumber,, globalOptArray(i).id
    put #filenumber,, cast(longint, len(globalOptArray(i).settingString))
    put #filenumber,, globalOptArray(i).settingString
    globalOptArray(i).fullList.save
  next
  close filenumber
end sub

function randomWord(words as TList = TList) as TWord
  ' returns a random word out of the list 'words'
  if words.list.isEmpty then return globallist.get(int(rnd*globallist.size))
  dim as mdList(TWord) list = randomWord(words, 1)
  if list.isEmpty then
    return TWord
  else
    return list.get(0)
  end if
end function

function randomWord(words as TList, byval count as integer) as mdList(TWord)
  ' returns a list of 'count' random words out of the list 'words'
  dim w as TWord, ret as mdList(TWord), list as mdList(long), lp as TListPoint, tempInt as integer
  dim as integer wordSize = words.list.size
  foreach(TListPoint, lp, words.list)
    if len(globallist.get(lp.id).word) then list.add(lp.id)
  nexteach
  if count > wordSize then count = wordSize
  for i as integer = wordSize to wordSize-count+1 step -1
    tempInt = int(rnd*i)
    w = globallist.get(list.get(tempInt))
    ret.add(w)
    list.remove(tempInt)
  next
  return ret
end function

function searchInput(txt as string = "", x as integer = SCREEN_X\2 - 250, y as integer = SCREEN_Y\2-150, _
                     w as integer = 500, h as integer = 300) as string
  ' messagebox to enter words with support of special keys
  dim as string key, ret = txt
  dim as integer mx, my, mb, cursor = len(txt)
  dim searchGui as TGui, event as TGuiEvent
  dim as SEARCHDIALOG.TData sdata
  wordToArray SPECIAL_KEYS, sdata.keyString(), 0
  SEARCHDIALOG.incgui searchGui, sdata
  dim as FB.Image ptr tempImg = imagecreate(sdata.w, sdata.h)
  x = (SCREEN_X - sdata.w)\2
  y = (SCREEN_Y - sdata.h)\2
  get (x, y)-step(sdata.w-1, sdata.h-1), tempImg
  line (x, y)-step(sdata.w-1, sdata.h-1), &hffffff, bf
  line (x+1, y+1)-step(sdata.w-3, sdata.h-3), 0, b
  line (x+2, y+2)-step(sdata.w-5, sdata.h-5), 0, b
  line (x+4, y+4)-step(sdata.w-9, sdata.h-9), 0, b
  dim as TInput inpt = TInput(x+sdata.inpX, y+sdata.inpY, FONTSIZE.SMALL, localize("input") & ": ", "", -1, sdata.inpW)
  screenlock
  searchGui.show
  inpt.show
  screenunlock
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  do
    key = inkey
    event = searchGui.getEvent(key)
    if len(key) orelse event.id > 2 then
      inpt.check(key, event.id, sdata.posKeys, sdata.keyString())
      screenlock
      inpt.show
      screenunlock
      if event.id > 2 then
        do
          getmouse mx, my,, mb
          sleep 1
        loop until mb < 1
      end if
    end if
    select case event.id
      case 1                         ' OK: return input
        ret = inpt.value
        exit do
      case 2                         ' Chancel: return old value
        ret = chr(27)
        exit do
    end select
    sleep 1
  loop until key = chr(255, 107)
  if key = chr(255, 107) then ret = key
  put (x, y), tempImg, pset
  imagedestroy tempImg
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  return ret
end function

function searchWord(word as string, padding as string = "") as TWord
  ' search the _first_ matching word in the global list
  ' You can look for a word (TWord.word; padding="word"), a translation (TWord.trans.get(X); padding="trans")
  ' or any map entry (TWord.map.get("mapKey"); padding="mapKey")
  dim w as TWord, i as long, ks as mdList(long) = globallist.keyset
  foreach(long, i, ks)
    w = globallist.get(i)
    select case padding
      case "trans"
        if w.translate.contains("word") then return w
      case "", "word"
        if w.word = word then return w
      case else
        if w.map.get(padding) = word then return w
    end select
  nexteach
  return TWord
end function

function searchWord(wordlist as TList, word as string, padding as string = "") as TWord
  ' search the _first_ matching word in a list
  ' You can look for a word (TWord.word; padding="word"), a translation (TWord.trans.get(X); padding="trans")
  ' or any map entry (TWord.map.get("mapKey"); padding="mapKey")
  dim l as TListPoint, w as TWord
  foreach(TListPoint, l, wordlist.list)
    w = globallist.get(l.id)
    select case padding
      case "trans"
        if w.translate.contains("word") then return w
      case "", "word"
        if w.word = word then return w
      case else
        if w.map.get(padding) = word then return w
    end select
  nexteach
  return TWord
end function

function standardInfo(exercise as TExercise) as string
  ' information string for the exercise (shown in the main window)
  dim as TList ptr listPtr = iif(exercise.currentList.size = 0, @exercise.fullList, @exercise.currentList)
  return localize("stdInfoTodo", str(listPtr->notChecked), str(listPtr->size))
end function

sub wordToArray(txt as string, arr() as string, opt as integer = 1)
  ' sort letters in 'txt'
  ' opt=0: do not sort
  '    =1: sort alphabetical
  '    =2: sort randomly (TODO)
  dim as integer utf8_1, utf8_2, asc1, asc2, txtSize = len(txt)
  dim as wstring ptr wTxt = callocate((txtSize+1)*sizeof(wstring))
  UTFToWChar(UTF_ENCOD_UTF8, strptr(txt), wTxt, @txtSize)

  for i as integer = 0 to len(*wTxt)-2
    if opt = 1 then
      utf8_1 = wTxt[i]
      select case as const utf8_1
        case 32 to 127          : asc1 = utf8_1
        case 193, 225, 196, 228 : asc1 =  97  ' Á á Ä ä
        case 201, 233           : asc1 = 101  ' É é
        case 205, 237           : asc1 = 105  ' Í í
        case 209, 241           : asc1 = 110  ' Ñ ñ
        case 211, 243, 214, 246 : asc1 = 111  ' Ó ó Ö ö
        case 223                : asc1 = 115  ' ß
        case 218, 250, 220, 252 : asc1 = 117  ' Ú ú Ü ü
        case 161, 191           : asc1 =  33  ' ¡ ¿
      end select
    end if
    if opt = 1 then
      for k as integer = i+1 to len(*wTxt)-1
        utf8_2 = wTxt[k]
        select case as const utf8_2
          case 32 to 127          : asc2 = utf8_2
          case 193, 225, 196, 228 : asc2 =  97  ' Á á Ä ä
          case 201, 233           : asc2 = 101  ' É é
          case 205, 237           : asc2 = 105  ' Í í
          case 209, 241           : asc2 = 110  ' Ñ ñ
          case 211, 243, 214, 246 : asc2 = 111  ' Ó ó Ö ö
          case 223                : asc2 = 115  ' ß
          case 218, 250, 220, 252 : asc2 = 117  ' Ú ú Ü ü
          case 161, 191           : asc2 = 33   ' ¡ ¿
        end select
        if asc1 > asc2 then
          wTxt[i] = utf8_2
          wTxt[k] = utf8_1
          asc1 = asc2 : utf8_1 = utf8_2
        end if
      next
    elseif opt = 2 then
     ' TODO
    end if
  next
  if len(*wTxt) then
    redim arr(len(*wTxt)-1)
    dim as zstring ptr zptr = allocate(3)
    for i as integer = 0 to len(*wTxt)-1
      *zptr = chr(0, 0, 0)
      WCharToUTF(UTF_ENCOD_UTF8, wTxt+i, 1, zptr, @txtSize)
      arr(i) = *zPtr
    next
    deallocate zPtr
  else
    erase arr
  end if
  deallocate wTxt
end sub

