' module: flashcard system with spaced repetition

namespace WORDTEST
  type TData
    as integer posLevels = 1, posRepeat = 10, posHint = 11, posShow = 12, posSpeak = 13, posHelp = 14
    as integer posInfo = 15, posEdit = 16, posKeys = 20, posLetters = 50, posLabelProgress = 999, inpX, inpY, inpW
    as integer letterX, letterY, letterDist, letterDistY, letterSize
    as string  keyString(any)
  end type
  declare function callback(exercise as TExercise) as long
  declare function info(exercise as TExercise) as string
  declare sub printLetters(letterArr() as string, letterUsed() as integer, _
                           x as integer, y as integer, dx as integer, dy as integer, _
                           maxLength as integer, size as integer = 14)
  declare function settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
  declare sub incGui(gui as TGui, wdata as TData)
end namespace

globalExercises.add(TExercise(LANGSHORT1 & "-" & LANGSHORT2, LANGTITLE1 & " -> " & LANGTITLE2, _
                              "default-" & LANGSHORT1 & "-" & LANGSHORT2,  20, _
                              @WORDTEST.callback, @wordtestInfo, @WORDTEST.settings))
globalExercises.add(TExercise(LANGSHORT2 & "-" & LANGSHORT1, LANGTITLE2 & " -> " & LANGTITLE1, _
                              "default-" & LANGSHORT2 & "-" & LANGSHORT1,  -20, _
                              @WORDTEST.callback, @wordtestInfo, @WORDTEST.settings))

function WORDTEST.callback(exercise as TExercise) as long
  ' vocabulary test (spaced repetition with SuperMemo algorithm SM-2)
  if exercise.currentList.list.isEmpty then
    message(localize("noWords"))
    return RETVAR.emptyList
  end if
  dim as TData wdata
  wordToArray SPECIAL_KEYS, wdata.keyString(), 0
  dim listpoint as TListPoint, w as TWord
  dim as integer solved, i = -1, hint, col, count = 0, countAll = 0, mx, my, mb, sum, showLetters, letterUsed()
  dim as string key, firstTrans, tempString, article, letters, letterArr(), entryArr()
  dim as FB.Image ptr okImg = imagecreate(16, 16)
  bload exepath & "/img/ok.bmp", okImg
  dim as TWord outWord

  ' prepare GUI
  dim testGui as TGui, tempButton as TGuiElement
  dim as TGuiElement labelProgress = TGuiElement(GuiElementType.label, wdata.posLabelProgress, _
                                     630, 600, 100, 20, "",,,, GuiState.noBorder)
  dim as TGuiEvent event
  incGui testGui, wdata
  testGui.elements.add labelProgress
  
  ' colors for the progress bar
  ' 0: rated as 'easy';  1: rated as 'difficult'; 2: to repeat;  3: to do
  dim as uinteger barCol(7) = { rgb(  0, 183, 0), rgb(  0, 255, 0), rgb(127, 255, 0), _
                                rgb(255, 255, 0), rgb(255, 183, 0), rgb(255, 127, 0), _
                                rgb(127, 127, 127), rgb(255, 255, 255) }

  ' How many words shall be asked?
  for k as integer = 0 to abs(exercise.opt)-1
    if exercise.currentList.list.get(k).position > now then exit for
    count += 1
  next
  showLetters = 0
  do
    dim as integer progress(7) = { 0, 0, 0, 0, 0, 0, 0, count }
    sum = count
    countAll += count

    do until progress(6) + progress(7) = 0
      tempButton = testGui.getElement(wdata.posInfo)
      tempButton.state or= GuiState.deactivated
      testGui.setElement(tempButton)
      tempButton.show
      if GlobalSettings and GLOBAL_SETTINGS.ALLOW_EDITS then
        tempButton = testGui.getElement(wdata.posEdit)
        tempButton.state or= GuiState.deactivated
        testGui.setElement(tempButton)
        tempButton.show
      end if
      if progress(6) + progress(7) = 1 then
        ' last word to ask; it's not possible to repeat the word any more
        tempButton = testGui.getElement(wdata.posRepeat)
        tempButton.state or= GuiState.deactivated
        testGui.setElement(tempButton)
        tempButton.show
      end if
      i += 1
      solved = 0
      listPoint = exercise.currentList.list.get(progress(6))
      w = globallist.get(listPoint.id)
      if len(w.word) = 0 then
        exercise.currentList.list.remove(progress(6))
        continue do
      end if
      dim as TInput inpt = TInput(wdata.inpX, wdata.inpY, FONTSIZE.NORMAL, localize("input") & ": ", "", -1, wdata.inpW)
      inpt.valueX = 190
      hint = 0
      letters = iif(exercise.opt > 0, w.translate.get(0), w.word)
      wordToArray letters, letterArr()
      redim letterUsed(ubound(letterArr))
      if exercise.opt > 0 then firstTrans = w.translate.get(0) else firstTrans = w.word
      for k as integer = 0 to ubound(wdata.keyString)
        tempButton = testGui.getElement(wdata.posKeys+k)
        tempButton.state and= not GuiState.deactivated
        testGui.setElement(tempButton)
      next
      tempButton = testGui.getElement(wdata.posHint)
      tempButton.state and= not GuiState.deactivated
      testGui.setElement(tempButton)
      labelProgress.title = countAll - progress(6) - progress(7) + 1 & " / " & countAll
      testGui.setElement(labelProgress)
      screenlock
      cls
      testGui.show
      drawProgress 200, 610, 400, 10, progress(), barCol()
      outWord = prettyOutput(w, iif(exercise.opt > 0, PrettOut.hint1 or PrettOut.obj, PrettOut.hint2))
      if exercise.opt > 0 then
        printFT 100, 80, rtrim(outWord.word), font
      else
        printFT 100, 80, outWord.translate.get(0), font
      end if
      inpt.show
      if showLetters then
        printLetters letterArr(), letterUsed(), wdata.letterX, wdata.letterY, _
                     wdata.letterDist, wdata.letterDistY, MAX_WORDLENGTH, wdata.letterSize
      end if
      screenunlock
      do
        getmouse mx, my,, mb
        sleep 1
      loop until mb < 1
      do
        key = inkey
        getmouse mx, my,, mb
        event = testGui.getEvent(key)
        if solved = 0 then
          inpt.check(key, event.id, wdata.posKeys, wdata.keyString())
          ' update list of letters
          if len(key) orelse event.id >= wdata.posKeys and event.id < wdata.posKeys+ubound(wdata.keyString)+2 then
            for i as integer = 0 to ubound(letterUsed)
              letterUsed(i) = 0
            next
            wordToArray inpt.value, entryArr(), 0
            for i as integer = 0 to ubound(entryArr)
              for k as integer = 0 to ubound(letterUsed)
                if letterUsed(k) = 0 andalso entryArr(i) = letterArr(k) then
                  letterUsed(k) = -1
                  exit for
                end if
              next
            next
            if showLetters then
              printLetters letterArr(), letterUsed(), wdata.letterX, wdata.letterY, _
                           wdata.letterDist, wdata.letterDistY, MAX_WORDLENGTH, wdata.letterSize
            end if
          end if
        end if
        select case event.id
        case wdata.posLetters
          showLetters = not showLetters
          if showLetters then
            printLetters letterArr(), letterUsed(), wdata.letterX, wdata.letterY, _
                         wdata.letterDist, wdata.letterDistY, MAX_WORDLENGTH, wdata.letterSize
          else
            line (wdata.letterX, wdata.letterY)-step(SCREEN_X-wdata.letterX, wdata.letterSize*2), &hffffff, bf
          end if
        case wdata.posLetters+1 to wdata.posLetters + ubound(letterArr)+1
          if showLetters andalso solved = 0 andalso letterUsed(event.id-wdata.posLetters-1) = 0 then
            letterUsed(event.id-wdata.posLetters-1) = -1
            inpt.value = left(inpt.value, inpt.cursor) & letterArr(event.id-wdata.posLetters-1) _
                         & mid(inpt.value, inpt.cursor+1)
            inpt.cursor += len(letterArr(event.id-wdata.posLetters-1))
            printLetters letterArr(), letterUsed(), wdata.letterX, wdata.letterY, _
                         wdata.letterDist, wdata.letterDistY, MAX_WORDLENGTH, wdata.letterSize
            if wordLength(inpt.value) = ubound(letterArr)+1 then
              for i as integer = 0 to ubound(letterUsed)
                if letterUsed(i) = 0 then exit select
              next
              solved = 1
            end if
          end if
        case wdata.posShow
          if solved = 0 then solved = 1
        case wdata.posSpeak
          if exercise.opt > 0 or solved = 2 then
            speakThread w.word
          end if
        case wdata.posRepeat
          progress(7) -= 1 : progress(6) += 1  ' move from 'to do' to 'to repeat'
          exit do
        case wdata.posHint
          if solved = 0 then
            hint += 1
            do until hint >= len(firstTrans) orelse firstTrans[hint] < 128 or firstTrans[hint] > 191
              hint += 1
            loop
            if hint >= len(firstTrans) then solved = 1
          end if
        case wdata.posInfo
          dim as FB.Image ptr tempImg = imagecreate(SCREEN_X, SCREEN_Y)
          get (0, 0)-(SCREEN_X-1, SCREEN_Y-1), tempImg
          wordInfo w
          put (0, 0), tempImg, pset
          imagedestroy tempImg
          do
            getmouse mx, my,, mb
            sleep 1
          loop until mb < 1
        case wdata.posEdit
          dim as FB.Image ptr tempImg = imagecreate(SCREEN_X, SCREEN_Y)
          get (0, 0)-(SCREEN_X-1, SCREEN_Y-1), tempImg
          EDITWORDS.edit w.id
          put (0, 0), tempImg, pset
          imagedestroy tempImg
          do
            getmouse mx, my,, mb
            sleep 1
          loop until mb < 1
        case wdata.posLevels to wdata.posLevels+5
          dim as TListPoint l
          dim as integer k, fulllistIndex = exercise.fullList.list.indexOf(listPoint)
          listPoint.modify(event.id-wdata.posLevels)
          listPoint.position = now + listPoint.addition - rnd/10
          listPoint.special += 1  ' works as counter
          ' update word level
          if exercise.opt > 0 then
            w.levelA = calculateLevel(listPoint.addition, listPoint.factor)
          else
            w.levelB = calculateLevel(listPoint.addition, listPoint.factor)
          end if
          globallist.put(listPoint.id, w)
          ' update progress bar
          progress(5 + wdata.posLevels - event.id) += 1
          progress(7) -= 1
          ' update position in exercise.fullList and exercise.currentList
          for lno as integer = 0 to 1
            dim as TList ptr listPtr = iif(lno=0, @exercise.fullList, @exercise.currentList)
            dim as integer index = iif(lno=0, fulllistIndex, progress(6))
            for k = index+1 to listPtr->size - 1
              l = listPtr->list.get(k)
              if l.position > listPoint.position then exit for
            next
            if k = listPtr->size then
              listPtr->list.add(listPoint)
            else
              listPtr->list.add(k, listPoint)
            end if
            listPtr->list.remove(index)
            listPtr->update
          next
          exit do
        case wdata.posHelp
          dim as integer ret = showHelpScreen(MODULE_PATH & "doc.htm", "simpleTest")
          if ret = RETVAR.windowClosed then key = chr(255, 107)
      end select
        if solved = 1 then
          for k as integer = 0 to ubound(wdata.keyString)
            tempButton = testGui.getElement(wdata.posKeys+k)
            tempButton.state or= GuiState.deactivated
            testGui.setElement(tempButton)
            tempButton.show
          next
          tempButton = testGui.getElement(wdata.posHint)
          tempButton.state or= GuiState.deactivated
          testGui.setElement(tempButton)
          tempButton.show
          tempButton = testGui.getElement(wdata.posInfo)
          tempButton.state = GuiState.noBorder
          testGui.setElement(tempButton)
          tempButton.show
          if GlobalSettings and GLOBAL_SETTINGS.ALLOW_EDITS then
            tempButton = testGui.getElement(wdata.posEdit)
            tempButton.state = GuiState.noBorder
            testGui.setElement(tempButton)
            tempButton.show
          end if
          ' special handling for nouns (colored output, article ...)
          outWord = prettyOutput(w, PRETTOUT.FULL, col)
          screenlock
          line (100, 145)-step(600, 0)
          line (90, 146)-step(700, 35), rgb(255, 255, 255), bf
          if exercise.opt > 0 then
            printFT 100, 150, localize("solution") & ":", font
            for k as integer = 0 to w.translate.size-1
              printFT 190, 150+25*k, w.translate.get(k), font
              if outWord.translate.get(k) = inpt.value then put (80, 150), okImg, pset
            next
            line (100, 80)-step(600, 30), rgb(255, 255, 255), bf
            printFT 100, 80, outWord.word, font,, col
          else
            printFT 100, 150, localize("solution") & ":", font
            printFT 190, 150, outWord.word, font,, col
            if outWord.word = inpt.value or w.word = inpt.value then put (80, 150), okImg, pset
          end if
          inpt.show(0)
          screenunlock
          solved = 2
        end if
        if solved = 0 andalso len(key) or event.id then
          screenlock
          inpt.show
          if hint > 0 then
            line (90, 146)-step(700, 35), rgb(255, 255, 255), bf
            printFT 100, 150, localize("hint") & ":", font,, &h777777
            printFT 190, 150, left(firstTrans, hint) & "...", font,, &h777777
          end if
          screenunlock
        end if
        if key = chr(255, 107) then exit do, do, do
        if key = chr(27) then
          select case message(localize("askQuitExerc"), localize("yes|no"), 2)
            case 1
              exit do, do, do
            case RETVAR.windowClosed
              key = chr(255, 107)
              exit do, do, do
          end select
        end if
        sleep 1
      loop
      if progress(7) = 0 then
        ' move all 'to repeat' to 'to do'
        progress(7) = progress(6)
        progress(6) = 0
      end if
    loop

    ' How many words shall be asked in the next step?
    count = 0
    for k as integer = 0 to abs(exercise.opt)-1
      if exercise.currentList.list.get(k).position > now then exit for
      count += 1
    next
    if sum then drawProgress 200, 610, 400, 10, progress(), barCol()
    if count = 0 then
      message(localize("noMoreWords"))
    else
      count *= (2-message(localize("askContinue", str(count)), localize("yes|no"), 2))
    end if
    if count then
      ' re-activate repeat button
      tempButton = testGui.getElement(wdata.posRepeat)
      tempButton.state and= not GuiState.deactivated
      testGui.setElement(tempButton)
      tempButton.show
    end if
  loop until count = 0

  imagedestroy okImg
  select case key
    case chr(255, 107) : return RETVAR.windowClosed
    case chr(27)       : return RETVAR.aborted
    case else          : return RETVAR.ok
  end select
  return RETVAR.ok
end function

sub WORDTEST.printLetters(letterArr() as string, letterUsed() as integer, _
                          x as integer, y as integer, dx as integer, dy as integer, _
                          maxLength as integer, size as integer = 14)
  dim as integer lx, ly
  screenlock
  line (x, y)-step(SCREEN_X-x, dy*2), &hffffff, bf
  for i as integer = 0 to ubound(letterArr)
    lx = x+(i mod ((maxLength-1)\2+1))*dx
    ly = iif(i > (maxLength-1)\2, y+dy, y)
    centerFT lx+.5*dx-1, ly+size+1, iif(letterArr(i) = " ", "_", letterArr(i)), font, FONTSIZE.X_SMALL, _
             iif(letterUsed(i), rgb(190, 190, 190), rgb(0, 0, 0))
    line (lx+2, ly+2)-step(dx-4, 2*size-4), iif(letterUsed(i), rgb(190, 190, 190), rgb(0, 0, 0)), b
  next
  screenunlock
end sub

function WORDTEST.settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
  dim as string settingSpecial = left(exercise.settingString, 8), settingWords = mid(exercise.settingString, 9)
  dim as integer ret = chooseWords(exercise, settingWords, optional or SETTING_OPTIONS.SHOWTAGS)
  exercise.settingString = settingSpecial & settingWords
  return ret
end function
