#ifndef PI
  const PI = 3.141593
#endif

namespace QUEHORA
  declare sub clock(t as double, x as integer, y as integer, w as integer, _
                    bgCol as uinteger = rgb(255, 255, 255), fgCol as uinteger = rgb(0, 0, 0), _
                    handCol as uinteger = rgb(32, 32, 32))  
  declare function callback(exercise as TExercise) as long
  declare function settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
end namespace

globalExercises.add(TExercise("quehora", "¿Que hora es?", "full", 0, @QUEHORA.callback, 0, 0,, "grammar"))

sub QUEHORA.clock(t as double, x as integer, y as integer, w as integer, _  ' time, geometry
                  bgCol as uinteger   = rgb(255, 255, 255), _               ' color of the face
                  fgCol as uinteger   = rgb(  0,   0,   0), _               ' color of the scale
                  handCol as uinteger = rgb( 32,  32,  32))                 ' color of the hands
  dim as FB.Image ptr img = imagecreate(w, w)
  dim as integer rIn, r = w\2                    ' radius and center of the clock
  dim as double hour_ = pi*t*4, min_ = pi*t*48   ' angle of the hands
  dim as double scalePos
  ' clock face
  circle img, (r, r), r-1, bgCol,,,, f
  circle img, (r, r), r-1, fgCol
  for i as integer = 1 to 60
    scalePos = pi/-30*i
    rIn = (r-1)*iif(i mod 5, .95, .8)
    line img, (r-sin(scalePos)*(r-1), r-cos(scalePos)*(r-1))-(r-sin(scalePos)*rIn, r-cos(scalePos)*rIn), fgCol
  next
  ' clock hands - hour
  line img, (r, r)-step(sin(hour_)*r*.5, -cos(hour_)*r*.5), handCol
  line img, (r-cos(hour_)*.5, r-sin(hour_)*.5)-step(sin(hour_)*r*.5, -cos(hour_)*r*.5), handCol
  line img, (r+cos(hour_)*.5, r+sin(hour_)*.5)-step(sin(hour_)*r*.5, -cos(hour_)*r*.5), handCol
  line img, (r-cos(hour_), r-sin(hour_))-step(sin(hour_)*r*.5, -cos(hour_)*r*.5), handCol
  line img, (r+cos(hour_), r+sin(hour_))-step(sin(hour_)*r*.5, -cos(hour_)*r*.5), handCol
  ' clock hands - minute
  line img, (r, r)-step(sin(min_)*r*.9, -cos(min_)*r*.9), handCol
  line img, (r-cos(min_)*.5, r-sin(min_)*.5)-step(sin(min_)*r*.9, -cos(min_)*r*.9), handCol
  line img, (r+cos(min_)*.5, r+sin(min_)*.5)-step(sin(min_)*r*.9, -cos(min_)*r*.9), handCol
  line img, (r-cos(min_), r-sin(min_))-step(sin(min_)*r*.9, -cos(min_)*r*.9), handCol
  line img, (r+cos(min_), r+sin(min_))-step(sin(min_)*r*.9, -cos(min_)*r*.9), handCol
  put (x, y), img, trans
  imagedestroy img
end sub

function QUEHORA.callback(exercise as TExercise) as long
  dim as integer MAXLENGTH = 35
  dim as TGui horaGui 
  dim as TGuiEvent event
  dim as integer hour_, hourMod, min_, cursor, mx, my, mb, count = 10
  dim as string key, specialKeys(), answer
  dim as string minName(6)   = { "en punto", "cinco", "diez", "cuarto", "veinte", "veinticinco", "media" }
  dim as string hourName(11) = { "doce", "una", "dos", "tres", "cuatro", "cinco", _
                                 "seis", "siete", "ocho", "nueve", "diez", "once" }
  enum
    posShow = 1, posHelp, posExit
    posKeys = 10
  end enum
  
  wordToArray SPECIAL_KEYS, specialKeys(), 0
  guiAddKeys horaGui, posKeys, specialKeys(), 40, 400, 30, 30, 15, 15, 11
    ' button: show correct answer

  dim as TGuiElement buttonShow = _
      TGuiElement(GuiElementType.button, posShow, 310, 530, 160, 50, "Antwort prüfen", chr(13),, "Enter")
  buttonShow.fontsize = 16
  horaGui.addElement buttonShow
  dim as TGuiElement buttonExit = _
      TGuiElement(GuiElementType.button, posExit, 580, 530, 160, 50, "Übung beenden", chr(27),, "ESC")
  buttonExit.fontsize = 16
  horaGui.addElement buttonExit
  horaGui.addElement _
      TGuiElement(GuiElementType.button, posHelp, _
                  exepath & "/img/help.bmp", 70, 540, 40, 40, chr(255, 59),, "F1", GuiState.noBorder)

  dim as uinteger barCol(2)  = { rgb(  0, 255, 0), rgb(255, 0, 0), rgb(255, 255, 255) }
  dim as integer progress(2) = { 0, 0, count }
  dim as TInput inpt = TInput(50, 140, FONTSIZE.NORMAL, localize("input") & ": ", "", -1, 435)

  do while progress(2) > 0
    cls
    horaGui.show
    hour_ = int(rnd*12)
    min_  = int(rnd*12)
    hourMod = iif(min_ > 6, (hour_+1) mod 12, hour_)
    answer  = iif(hourMod = 1, "Es la ", "Son las ") & hourName(hourMod) & " "
    if min_ = 0 then
      answer &= minName(0)
    elseif min_ < 7 then
      answer &= "y " & minName(min_)
    else
      answer &= "menos " & minName(12-min_)
    end if
    inpt.value = ""
    printFT 50, 100, "¿Qué hora es?", font, 18
    clock timeserial(hour_, min_*5, 0), 500, 100, 200
    inpt.show
    drawProgress 200, 610, 400, 10, progress(), barCol()
    do
      getmouse mx, mx,, mb
      sleep 1
    loop until mb < 1
    do
      key = inkey
      event = horaGui.getEvent(key)
      if len(key) orelse event.id then
        inpt.check(key, event.id, posKeys, specialKeys())
        inpt.show
      end if
      select case event.id
        case posHelp
          dim as integer ret = showHelpScreen(MODULE_PATH & "doc.htm", "quehora")
          if ret = RETVAR.windowClosed then key = chr(255, 107)
        case posExit
          select case message(localize("askQuitExerc"), localize("yes|no"), 2)
            case 1                   : exit do
            case RETVAR.windowClosed : key = chr(255, 107)
          end select
      end select
      sleep 1
    loop until event.id = posShow or key = chr(255, 107)
    if key = chr(255, 107) then return RETVAR.windowClosed
    if event.id = posExit then return RETVAR.aborted

    ' check input
    if lcase(inpt.value) = lcase(answer) or lcase(inpt.value) = lcase(answer) & "." then
      messageButton rgb(0, 255, 0), rgb(0, 0, 0), "¡Es verdad!", 20, 40, 300, 410, 50
      progress(0) += 1
    else
      messageButton rgb(255, 63, 63), rgb(0, 0, 0), "¡Falso! " & answer & ".", 20, 40, 300, 410, 50
      progress(1) += 1
    end if
    progress(2) -= 1
    drawProgress 200, 610, 400, 10, progress(), barCol()

    buttonShow.title = "Weiter"
    buttonShow.show
    do
      getmouse mx, mx,, mb
      sleep 1
    loop until mb < 1
    do
      key = inkey
      event = horaGui.getEvent(key)
      select case event.id
        case posHelp
          dim as integer ret = showHelpScreen(MODULE_PATH & "doc.htm", "quehora")
          if ret = RETVAR.windowClosed then key = chr(255, 107)
        case posExit
          if progress(2) < 1 then event.id = posShow : exit do
          select case message(localize("askQuitExerc"), localize("yes|no"), 2)
            case 1                   : exit do
            case RETVAR.windowClosed : key = chr(255, 107)
          end select
      end select
      sleep 1
    loop until event.id = posShow or key = chr(255, 107)
    if key = chr(255, 107) then return RETVAR.windowClosed
    if event.id = posExit then return RETVAR.aborted
    if key = chr(255, 107) then return RETVAR.windowClosed
    messageButton -1
    buttonShow.title = localize("continue")
    buttonShow.show
  loop
  return RETVAR.ok
end function

function QUEHORA.settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
  return -1
end function

