function TPageStandardOutput(x as integer, y as integer, x2 as integer, h as integer, _
                             word as TWord, opt as integer) as integer
  dim as integer ret = word.translate.size*h
  if opt then
    printFT x, y, word.word, font, opt
    for i as integer = 0 to ret-1
      printFT x2, y+i*h, word.translate.get(i), font, opt
    next
  end if
  return ret
end function

type TPageEntry
  as long id, x, y, h
  declare constructor
  declare constructor(id as integer, x as integer, y as integer, h as integer)
  declare operator cast as string
end type

constructor TPageEntry
end constructor
constructor TPageEntry(id as integer, x as integer, y as integer, h as integer)
  this.id = id
  this.x  = x
  this.y  = y
  this.h  = h
end constructor
operator TPageEntry.cast as string
  return cvl(id) & cvl(x) & cvl(y) & cvl(h)
end operator
operator =(e1 as TPageEntry, e2 as TPageEntry) as integer
  return e1.id = e2.id
end operator

mdListDeclare(TPageEntry)
mdListDeclare(mdList(TPageEntry))

type TPages
  as mdList(mdList(TPageEntry)) pages
  as integer fSize, startX, startY, wordH, wordDist, transDist, colH, colDist, cols
  as integer colWidth, wordWidth, transWidth, barPos, barState
  as function (x as integer, y as integer, xPlus as integer, h as integer, _
               word as TWord, fontSize as integer) as integer outp
  dim as TPageEntry wordPos(any)
  declare constructor
  declare constructor(fSize as integer, startX as integer, startY as integer, wordH as integer, wordDist as integer, _
                      transDist as integer, colH as integer, colDist as integer = 0, cols as integer = 1)
  declare sub calcPositions(words as mdList(TListPoint))
  declare sub calcPositions(words as TList)
  declare sub printEntries(words as mdList(TListPoint), page as integer = 0)
  declare sub printEntries(words as TList, page as integer = 0)
  declare sub showBar
  declare sub moveBar(p as integer)
end type

constructor TPages
  barState = 1
  outp     = @TPageStandardOutput
end constructor
constructor TPages(fSize as integer, startX as integer, startY as integer, wordH as integer, wordDist as integer, _
                   transDist as integer, colH as integer, colDist as integer = 0, cols as integer = 1)
  this.fSize      = fSize
  this.startX     = startX
  this.startY     = startY
  this.wordH      = wordH
  this.wordDist   = wordDist
  this.transDist  = transDist
  this.colH       = colH
  this.colDist    = iif(colDist=0, SCREEN_X-2*startX, colDist)
  this.cols       = cols
  this.barPos     = 0
  this.barState   = 1
  this.outp       = @TPageStandardOutput
  this.wordWidth  = transDist
  this.transWidth = transDist
  this.colWidth   = 2*transDist
end constructor

sub TPages.calcPositions(words as mdList(TListPoint))
  dim w as TWord, listPoint as TListPoint, curPage as mdList(TPageEntry)
  dim as integer posX = 0, posY = 0, h
  dim as mdList(TPageEntry) pageList

  foreach(mdList(TPageEntry), pageList, pages)
    pageList.clear
  nexteach
  pages.clear
  foreach(TListPoint, listPoint, words)
    w = globallist.get(listPoint.id)
    h = outp(0, 0, 0, wordH, w, 0)
    if posY + h > colH then
      if posX >= colDist*(cols-1) then
        posX = 0 : posY = 0
        pages.add(curPage)
        curPage.clear
      else
        posX += colDist : posY = 0
      end if
    else
    end if
    curPage.add(TPageEntry(w.id, posX+startX, posY+startY, h))
    posY += h + wordDist
  nexteach
  pages.add(curPage)
end sub

sub TPages.calcPositions(words as TList)
  calcPositions words.list
end sub

sub TPages.printEntries(words as mdList(TListPoint), page as integer = 0)
  dim w as TWord, e as TPageEntry, curPage as mdList(TPageEntry) = pages.get(page), i as integer, h as integer
  if curPage.size then
    redim wordPos(curPage.size-1)
    foreach(TPageEntry, e, curPage)
      wordPos(i) = e
      w = globallist.get(e.id)
      h = outp(e.x, e.y, e.x+transDist, wordH, w, fSize)
      line (e.x, e.y+h+wordDist\2)-step(colWidth-10, 0), &hcccccc
      i += 1
    nexteach
  end if
end sub

sub TPages.printEntries(words as TList, page as integer = 0)
  printEntries words.list, page
end sub

sub TPages.showBar
  if barState = 0 then exit sub
  dim as integer x = iif(barPos <= ubound(wordPos), wordPos(barPos).x, wordPos(ubound(wordPos)).x)
  dim as integer xPlus = iif(barState = 1, transDist, 0)
  dim as integer w = iif(barState = 1, transWidth, wordWidth)-10
  dim as integer y = iif(barPos <= ubound(wordPos), wordPos(barPos).y, _
                         wordPos(ubound(wordPos)).y + wordPos(ubound(wordPos)).h + wordDist) 
  line (x+xPlus, y)-(x+xPlus+w, startY+colH+wordDist), &h999999, bf
  for i as integer = x-startX+colDist to wordPos(ubound(wordPos)).x step colDist
    line (startX+i+xPlus, startY)-step(w, colH+wordDist), &h999999, bf
  next
end sub

sub TPages.moveBar(p as integer)
  barPos += p
  if barPos < 0 then
    barPos = 0
  elseif barPos > ubound(wordPos)+1 then
    barPos = ubound(wordPos)+1
  end if
end sub
